SET NOCOUNT ON;
go
USE ImEx;
go
IF EXISTS
 (SELECT *
  FROM   sys.objects
  WHERE  object_id = OBJECT_ID(N'dbo.SetBasedIter')
  AND    type IN (N'P', N'PC'))
BEGIN;
  DROP PROCEDURE dbo.SetBasedIter;
END;
go
CREATE PROCEDURE dbo.SetBasedIter
                 @MaxCandidatesPerSession smallint
AS
-- Create temporary table for efficient update of SpaceLeft in dbo.Sessions
CREATE TABLE #SessionsAssigned
            (Year          smallint NOT NULL
            ,Quarter       tinyint  NOT NULL
            ,SessionNo     smallint NOT NULL
            ,NumCandidates smallint NOT NULL
            ,CONSTRAINT PK_SessionsAssigned
                 PRIMARY KEY (Year, Quarter, SessionNo)
            );

-- Keep adding and filling sessions until all entries are assigned to a session
WHILE 1 = 1
BEGIN;

  -- For each quarter, compute the minimal number of sessions needed
  --  for the as of yet unallotted registrations, using this formula:
  --    SUM(NumCandidates) / @MaxCandidatesPerSession (rounded up)
  -- Insert that number of (empty) sessions in the table.
  WITH SessionsNeeded
  AS (SELECT   Year, Quarter,
               CEILING(1.0 * SUM(NumCandidates) / @MaxCandidatesPerSession) AS Needed
      FROM     dbo.Registrations
      WHERE    SessionNo IS NULL         -- Exlude registrations that are already allocated
      GROUP BY Year, Quarter)
 ,MaxSession
  AS (SELECT   Year, Quarter,
               MAX(SessionNo) AS MaxSession
      FROM     dbo.Sessions
      GROUP BY Year, Quarter)
  INSERT INTO  dbo.Sessions (Year, Quarter, SessionNo, SpaceLeft)
  SELECT       sn.Year,
               sn.Quarter,
               COALESCE(ms.MaxSession, 0) + n.Number,
               @MaxCandidatesPerSession
  FROM         SessionsNeeded AS sn
  INNER JOIN   dbo.Numbers    AS n
        ON     n.Number  BETWEEN 1 AND sn.Needed
  LEFT  JOIN   MaxSession     AS ms
        ON     ms.Year         = sn.Year
        AND    ms.Quarter      = sn.Quarter;

  -- No new sessions means that there are no more entries with SessionNo NULL - we're finished!
  IF @@ROWCOUNT = 0
    BREAK;

  -- Keep adding registrations to these sessions until they are full
  WHILE 1 = 1
  BEGIN;
    -- Clear out the temp table used for updating SpaceLeft after assigning registrations to sessions
    TRUNCATE TABLE #SessionsAssigned;
    
    -- Add another entry to each session that still has room for one.
    -- Note that this is not a best fit algorithm, but rather a compromise 
    -- to keep the query simple (cough!) and limit execution time.
    -- Each session gets a ranking based on descending SpaceLeft (within quarter).
    -- Each registration gets a ranking, based on descending NumberOfCandidates (within quarter).
    -- Entries that won't fit any available session are omitted - they'll wait for the next round of sessions.
    -- The tables are joined on rank and if the NumberOfCandidates doesn't exceed the space left, it is added.

    -- Note that T-SQL proprietary "UPDATE ... FROM" syntax is used.
    -- An equivalent ANSI-compliant UPDATE would require me to duplicate a subquery,
    -- which would have a huge negative impact on performance.
    WITH Threshold
    AS (SELECT     Year, Quarter,
                   MAX(SpaceLeft)      AS Threshold
        FROM       dbo.Sessions
        GROUP BY   Year, Quarter)
   ,RankedSessions
    AS (SELECT     Year, Quarter, SessionNo, SpaceLeft,
                   ROW_NUMBER()     OVER (PARTITION BY Year, Quarter
                                          ORDER BY SpaceLeft DESC)
                                       AS Ranking
        FROM       dbo.Sessions
        WHERE      SpaceLeft            > 0)
   ,RankedRegistrations
    AS (SELECT     r.Year, r.Quarter, r.NumCandidates, r.SessionNo,
                   ROW_NUMBER()     OVER (PARTITION BY r.Year, r.Quarter
                                          ORDER BY r.NumCandidates DESC)
                                       AS Ranking
        FROM       dbo.Registrations   AS r
        INNER JOIN Threshold           AS t
              ON   t.Year               = r.Year
              AND  t.Quarter            = r.Quarter
        WHERE      r.SessionNo         IS NULL
        AND        r.NumCandidates     <= t.Threshold)
    UPDATE         r
    SET            SessionNo            = s.SessionNo
    OUTPUT         inserted.Year,         inserted.Quarter,
                   inserted.SessionNo,    inserted.NumCandidates
    INTO           #SessionsAssigned (Year, Quarter, SessionNo, NumCandidates)
    FROM           RankedRegistrations AS r
    INNER JOIN     RankedSessions      AS s
          ON       s.Year               = r.Year
          AND      s.Quarter            = r.Quarter
          AND      s.Ranking            = r.Ranking
    WHERE          s.SpaceLeft         >= r.NumCandidates;

    -- No updates done means that no session has room for a registration, or all registrations are handled
    IF @@ROWCOUNT = 0
      BREAK;

    -- One or more registrations were assigned to sessions.
    -- Subtract their size from remaining room in those sessions.
    -- (Without the temporary table, all sessions would have to have their SpaceLeft recomputed!)
    -- Again using T-SQL proprietary UPDATE instead of ANSI standard for performance.
    UPDATE         s
    SET            s.SpaceLeft          = s.SpaceLeft - sa.NumCandidates
    FROM           dbo.Sessions        AS s
    INNER JOIN     #SessionsAssigned   AS sa
          ON       sa.Year              = s.Year
          AND      sa.Quarter           = s.Quarter
          AND      sa.SessionNo         = s.SessionNo;

  END; -- End inner loop (adding registrations to sessions)

END; -- End outer loop (adding sessions to be filled in inner loop)
go
